"""Predator-prey environment state generator"""
import abc

import chex
import jax.random
from jumanji.environments.swarms.common.types import AgentParams

from floxs.common.utils import random_agent_state

from .types import State


class Generator(abc.ABC):
    def __init__(
        self, num_predators: int, num_prey: int, env_size: float = 1.0
    ) -> None:
        """
        State generator abstract class

        Parameters
        ----------
        num_predators
            Number of predator agents
        num_prey
            Number of prey agents
        env_size
            Environment size, default 1.0
        """
        self.num_predators = num_predators
        self.num_prey = num_prey
        self.env_size = env_size

    @abc.abstractmethod
    def __call__(
        self, key: chex.PRNGKey, predator_params: AgentParams, prey_params: AgentParams
    ) -> State:
        """Generate initial agent positions and velocities

        Parameters
        ----------
        key
            JAX random key
        predator_params
            Predator agent parameters
        prey_params
            Prey agent parameters

        Returns
        -------
        State
            Initial environment state
        """


class RandomGenerator(Generator):
    def __call__(
        self, key: chex.PRNGKey, predator_params: AgentParams, prey_params: AgentParams
    ) -> State:
        """
        Initial state generator that random samples agent states

        Random generator that samples predator and prey agent positions
        and velocities from a uniform distribution

        Parameters
        ----------
        key
            JAX random key
        predator_params
            Predator agent parameters
        prey_params
            Prey agent parameters

        Returns
        -------
        State
            Randomly sampled initial environment state
        """
        k_predator, k_prey = jax.random.split(key)
        predator_states = random_agent_state(
            k_predator, predator_params, self.num_predators, self.env_size
        )
        prey_states = random_agent_state(
            k_prey, prey_params, self.num_prey, self.env_size
        )
        state = State(
            predators=predator_states,
            prey=prey_states,
            step=0,
        )
        return state



"""
Multi-agent predator-prey environment

Multi-agent environment where a Flock of predator
type agents attempt to capture a flock of prey
agents.
"""
from . import env, generator, observations, rewards, types, viewer
from .env import PredatorPrey



"""Predator-prey environment data types"""
from typing import TYPE_CHECKING, NamedTuple

import chex
from jumanji.environments.swarms.common.types import AgentState

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from chex import dataclass


@dataclass
class Actions:
    """
    Predator-prey actions

    predators
        Array of predator actions
    prey
        Array of prey actions
    """

    predators: chex.Array  # (num_predators, 2)
    prey: chex.Array  # (num_prey, 2)


@dataclass
class Rewards:
    """
    Predator-prey rewards

    predators
        Array of predator rewards
    prey
        Array of prey rewards
    """

    predators: chex.Array  # (num_predators,)
    prey: chex.Array  # (num_prey,)


@dataclass
class Discounts:
    """
    Predator-prey discounts

    predators
        Array of predator discounts
    prey
        Array of prey discounts
    """

    predators: chex.Array  # (num_predators,)
    prey: chex.Array  # (num_prey,)


@dataclass
class State:
    """
    Predator-prey environment state

    predators
        Predator agent states
    prey
        Prey agent states
    step
        Simulation step
    """

    predators: AgentState
    prey: AgentState
    step: int = 0


class Observation(NamedTuple):
    """
    Predator-prey observations

    predator_views
        Array of individual predator agent views
    prey_views
        Array of individual prey agent views
    step
        Simulation step
    predator_positions
        Array of predator agent positions in the environment
    prey_positions
        Array of prey agent positions in the environment
    """

    predator_views: chex.Array  # (num_predators, 2, num_vision)
    prey_views: chex.Array  # (num_prey, 2, num_vision)
    step: chex.Numeric  # ()
    predator_positions: chex.Array  # (num_predators, 2)
    prey_positions: chex.Array  # (num_prey, 2)



"""Common utilities"""
import chex
import jax
import jax.numpy as jnp
from jumanji.environments.swarms.common.types import AgentParams, AgentState


def random_agent_state(
    key: chex.PRNGKey, agent_params: AgentParams, n_agents: int, env_size: float
) -> AgentState:
    """
    Initialise a random agent state

    Sample agent position, heading, and speed from a uniform distribution.

    Parameters
    ----------
    key
        JAX random key
    agent_params
        Agent parameters
    n_agents
        Number of agents to generate
    env_size
        Size of the environment

    Returns
    -------
    AgentState
        Random agent states
    """
    k_pos, k_head, k_speed = jax.random.split(key, 3)
    positions = jax.random.uniform(k_pos, (n_agents, 2), minval=0.0, maxval=env_size)
    headings = jax.random.uniform(k_head, (n_agents,), minval=0.0, maxval=2.0 * jnp.pi)
    speeds = jax.random.uniform(
        k_speed,
        (n_agents,),
        minval=agent_params.min_speed,
        maxval=agent_params.max_speed,
    )
    return AgentState(
        pos=positions,
        speed=speeds,
        heading=headings,
    )



"""
Flock multi-agent RL environment

Multi-agent environment inspired by Reynolds boid model. Agents
aim to stay close to other agents without colliding.
"""
from functools import cached_property, partial
from typing import Optional, Sequence

import chex
import jax
import jax.numpy as jnp
from jumanji import specs
from jumanji.env import Environment
from jumanji.environments.swarms.common.types import AgentParams
from jumanji.environments.swarms.common.updates import update_state
from jumanji.types import TimeStep, restart, termination, transition
from jumanji.viewer import Viewer
from matplotlib.animation import FuncAnimation

from .generator import Generator, RandomGenerator
from .observations import ObservationFn
from .rewards import ExponentialRewardFn, RewardFn
from .types import Observation, State
from .viewer import FlockViewer


class Flock(Environment):
    def __init__(
        self,
        boid_max_rotate: float = 0.025,
        boid_max_accelerate: float = 0.001,
        boid_min_speed: float = 0.015,
        boid_max_speed: float = 0.025,
        time_limit: int = 400,
        boid_radius: float = 0.01,
        viewer: Optional[Viewer[State]] = None,
        generator: Optional[Generator] = None,
        reward_fn: Optional[RewardFn] = None,
        observation: Optional[ObservationFn] = None,
    ) -> None:
        """
        Flock multi-agent environment

        Multi-agent environment inspired by Reynolds boid model. Agents
        aim to stay close to other agents without colliding.

        Actions
        -------
        Each individual agent can rotate their heading, and change their
        speed. Actions are in the range ``[-1, 1]`` which are then
        scales by the maximum rotation and acceleration parameters. The
        total actions for the flock is an array with shape ``[n_agents, 2]``.
        The action values represent ``[rotation, acceleration]`` respectively.

        Observations
        ------------
        Agents individually observe their local environment within a
        given range. The view-cone of each agent is segmented into individual
        values, that represent the distance to a neighbouring agents along a
        ray. The default value is ``-1`` in the case no neighbour lies
        along that ray.

        Rewards
        -------
        Agents are rewarded for each agent within neighbouring range, but
        penalised for each agent they collide with. By default, rewards
        decrease exponentially with distance between neighbouring agents.

        Parameters
        ----------
        boid_max_rotate
            Max agent change in heading per step, as a fraction of π
        boid_max_accelerate
            Max agent change of speed per step.
        boid_min_speed
            Minimim agent speed
        boid_max_speed
            Maximum agent speed
        time_limit
            Environment time limit
        boid_radius
            Radius of agents for collision detection
        viewer
            Environment viewer used for visualisation, default
            uses a matplotlib backend
        generator
            Initial state generator, the default uniformly
            generated agent positions and velocities
        reward_fn
            Agent reward function, default rewards decay
            exponentially based on distance between agents
        observation
            Agent observation function, default is a standard
            segmented view showing the distance to neighbouring
            agents, or -1 is no agent is in range
        """
        self.boid_params = AgentParams(
            max_rotate=boid_max_rotate,
            max_accelerate=boid_max_accelerate,
            min_speed=boid_min_speed,
            max_speed=boid_max_speed,
        )
        self.time_limit = time_limit
        self.generator = generator or RandomGenerator(num_boids=50)
        self._viewer = viewer or FlockViewer()
        self._reward_fn = reward_fn or ExponentialRewardFn(
            boid_radius=boid_radius,
            collision_penalty=0.1,
            i_range=0.2,
        )
        self._observation_fn = observation or ObservationFn(
            num_vision=128,
            vision_range=0.2,
            view_angle=0.5,
            env_size=self.generator.env_size,
            boid_radius=boid_radius,
        )
        super().__init__()

    def __repr__(self) -> str:
        return "\n".join(
            [
                "Boid flock multi-agent environment:",
                f" - num boids: {self.generator.num_boids}",
                f" - boid max rotation: {self.boid_params.max_rotate}",
                f" - boid max acceleration: {self.boid_params.max_accelerate}",
                f" - boid min speed: {self.boid_params.min_speed}",
                f" - boid max speed: {self.boid_params.max_speed}",
                f" - boid vision range: {self._observation_fn.vision_range}",
                f" - boid view angle: {self._observation_fn.view_angle}",
                f" - num vision: {self._observation_fn.num_vision}",
                f" - boid radius: {self._observation_fn.boid_radius}",
                f" - time limit: {self.time_limit},"
                f" - env size: {self.generator.env_size}"
                f" - generator: {self.generator.__class__.__name__}",
                f" - reward fn: {self._reward_fn.__class__.__name__}",
                f" - observation fn: {self._observation_fn.__class__.__name__}",
            ]
        )

    def reset(self, key: chex.PRNGKey) -> tuple[State, TimeStep[Observation]]:
        """
        Reset the environment

        Generates a new initial environment state and initial timestep

        Parameters
        ----------
        key
            JAX random key

        Returns
        -------
        State, TimeStep
            New initial state and initial timestep
        """
        state = self.generator(key, self.boid_params)
        timestep = restart(
            observation=self._state_to_observation(state), shape=(self.num_agents,)
        )
        return state, timestep

    def step(
        self, state: State, actions: chex.Array
    ) -> tuple[State, TimeStep[Observation]]:
        """
        Apply actions and update the state of the environment

        The update performs several steps:

        - Apply the rotation and acceleration actions to each agent
        - Update the position of all the agents
        - Calculate rewards between all agents
        - Generate new observation for each individual agent

        Parameters
        ----------
        state
            Current environment state
        actions
            Array of individual agent actions, in the shape
            ``[n-agents, 2]`` giving the change in heading and
            change in speed to apply to each agent in the range
            ``[-1, 1]``. The values represent ``[rotation, acceleration]``
            respectively.

        Returns
        -------
        State, TimeStep
            Update state, and new timestep
        """
        boids = update_state(
            self.generator.env_size, self.boid_params, state.boids, actions
        )
        state = State(boids=boids, step=state.step + 1)
        rewards = self._reward_fn(state)
        observation = self._state_to_observation(state)
        observation = jax.lax.stop_gradient(observation)
        timestep = jax.lax.cond(
            state.step >= self.time_limit,
            partial(termination, shape=(self.num_agents,)),
            partial(transition, shape=(self.num_agents,)),
            rewards,
            observation,
        )
        return state, timestep

    def _state_to_observation(self, state: State) -> Observation:
        boid_views = self._observation_fn(state)
        return Observation(
            views=boid_views,
            step=state.step,
            positions=state.boids.pos / self.generator.env_size,
        )

    @cached_property
    def num_agents(self) -> int:
        """
        Number of agents in the flock

        Returns
        -------
        int
            Number of agents
        """
        return self.generator.num_boids

    @cached_property
    def observation_spec(self) -> specs.Spec[Observation]:
        """Returns the observation spec

        Each agent individually generates a segmented view
        showing the distance to neighbouring agents,
        or ``-1`` in the case no agent is present in view-range.

        Returns
        -------
        Spec
            Flock environment observation spec
        """
        boid_views = specs.BoundedArray(
            shape=(
                self.num_agents,
                self._observation_fn.num_vision,
            ),
            minimum=-1.0,
            maximum=1.0,
            dtype=float,
            name="boid_views",
        )
        return specs.Spec(
            Observation,
            "ObservationSpec",
            views=boid_views,
            step=specs.BoundedArray(
                shape=(),
                minimum=0,
                maximum=self.time_limit,
                name="step",
                dtype=jnp.int32,
            ),
            positions=specs.BoundedArray(
                shape=(self.num_agents, 2),
                minimum=0.0,
                maximum=1.0,
                name="positions",
                dtype=jnp.float32,
            ),
        )

    @cached_property
    def action_spec(self) -> specs.BoundedArray:
        """Returns the action spec

        2d array of individual agent actions. Each agents action is
        an array representing ``[rotation, acceleration]`` in the range
        ``[-1, 1]``.

        Returns
        -------
        BoundedArray
            Action array spec
        """
        return specs.BoundedArray(
            shape=(self.generator.num_boids, 2),
            minimum=-1.0,
            maximum=1.0,
            dtype=float,
        )

    @cached_property
    def reward_spec(self) -> specs.BoundedArray:
        """Returns the reward spec

        Array of individual rewards for each agent

        Returns
        -------
        BoundedArray
            Reward array spec
        """
        return specs.BoundedArray(
            shape=(self.generator.num_boids,),
            minimum=0.0,
            maximum=1.0,
            dtype=float,
        )

    def render(self, state: State) -> None:
        """Render a frame of the environment for a given state using matplotlib.

        Parameters
        ----------
        state
            State object
        """
        self._viewer.render(state)

    def animate(
        self,
        states: Sequence[State],
        interval: int = 100,
        save_path: Optional[str] = None,
    ) -> FuncAnimation:
        """Create an animation from a sequence of environment states

        Parameters
        ----------
        states
            Sequence of environment states corresponding to consecutive
            timesteps
        interval
            Delay between frames in milliseconds
        save_path
            The path where the animation file should be saved. If
            ``None``, the plot will not be saved

        Returns
        -------
        FuncAnimation
            Animation that can be saved as a GIF, MP4, or rendered with HTML
        """
        return self._viewer.animate(states, interval=interval, save_path=save_path)

    def close(self) -> None:
        """Perform any necessary cleanup."""
        self._viewer.close()



"""
Flock mult-agent environment

Multi-agent environment inspired by Reynolds boid model. Agents
aim to stay close to other agents without colliding.
"""
from floxs.flock import env, generator, observations, rewards, types, viewer
from floxs.flock.env import Flock



"""Flock environment visualisation utilities"""
from typing import Optional, Sequence, Tuple

import jax.numpy as jnp
import matplotlib.animation
import matplotlib.pyplot as plt
from jumanji.environments.swarms.common.viewer import draw_agents, format_plot
from jumanji.viewer import MatplotlibViewer
from matplotlib.artist import Artist
from numpy.typing import NDArray

from .types import State


class FlockViewer(MatplotlibViewer[State]):
    def __init__(
        self,
        name: str = "flock",
        env_size: Tuple[float, float] = (1.0, 1.0),
        boid_color: str = "#282B28",  # black
        render_mode: str = "human",
    ) -> None:
        """
        Flock environment visualiser using Matplotlib

        Parameters
        ----------
        name
            Plot name, default ``flock``
        env_size
            Tuple containing the dimensions of the environment
        boid_color
            Color applied to boid agents, default is black
        render_mode
            Default ``human``
        """
        self.env_size = env_size
        self.boid_color = boid_color
        super().__init__(name, render_mode)

    def render(
        self, state: State, save_path: Optional[str] = None
    ) -> Optional[NDArray]:
        """Render a frame of the environment for a given state using matplotlib

        Parameters
        ----------
        state
            State object containing the current dynamics of the environment
        save_path
            Optional path to save the rendered environment image to

        Returns
        -------
        Array or None
            RGB array if the render_mode is ``rgb_array``
        """
        self._clear_display()
        fig, ax = self._get_fig_ax()
        self._draw(ax, state)

        if save_path:
            fig.savefig(save_path, bbox_inches="tight", pad_inches=0.2)

        return self._display(fig)

    def animate(
        self, states: Sequence[State], interval: int, save_path: Optional[str] = None
    ) -> matplotlib.animation.FuncAnimation:
        """Create an animation from a sequence of states

        Parameters
        ----------
        states
            Sequence of ``State`` corresponding to subsequent timesteps
        interval
            Delay between frames in milliseconds, default to 200
        save_path
            The path where the animation file should be saved. If it is None,
            the plot will not be saved

        Returns
        -------
        FuncAnimation
            Animation object that can be saved as a GIF, MP4, or rendered with HTML
        """
        if not states:
            raise ValueError(f"The states argument has to be non-empty, got {states}.")
        fig, ax = self._get_fig_ax(name_suffix="_animation", show=False)
        fig, ax = format_plot(fig, ax, self.env_size)
        plt.close(fig=fig)

        boid_quiver = draw_agents(ax, states[0].boids, self.boid_color)

        def make_frame(state: State) -> tuple[Artist]:
            boid_quiver.set_offsets(state.boids.pos)
            boid_quiver.set_UVC(
                jnp.cos(state.boids.heading), jnp.sin(state.boids.heading)
            )
            return (boid_quiver,)

        self._animation = matplotlib.animation.FuncAnimation(
            fig,
            make_frame,
            frames=states,
            interval=interval,
            blit=True,
        )

        if save_path:
            self._animation.save(save_path)

        return self._animation

    def _draw(self, ax: plt.Axes, state: State) -> None:
        ax.clear()
        draw_agents(ax, state.boids, self.boid_color)

    def _get_fig_ax(
        self,
        name_suffix: Optional[str] = None,
        show: bool = True,
        padding: float = 0.05,
        **fig_kwargs: str,
    ) -> Tuple[plt.Figure, plt.Axes]:
        fig, ax = super()._get_fig_ax(
            name_suffix=name_suffix, show=show, padding=padding
        )
        fig, ax = format_plot(fig, ax, self.env_size)
        return fig, ax



"""Predator-prey environment observation function"""
from typing import Tuple

import chex
import esquilax
import jax.numpy as jnp
from jumanji.environments.swarms.common.updates import view, view_reduction_fn

from .types import State


def _view_reduction(view_shape: Tuple[int, ...]) -> esquilax.reductions.Reduction:
    return esquilax.reductions.Reduction(
        fn=view_reduction_fn,
        id=-jnp.ones(view_shape),
    )


class ObservationFn:
    def __init__(
        self,
        num_vision: int,
        predator_vision_range: float,
        prey_vision_range: float,
        predator_view_angle: float,
        prey_view_angle: float,
        agent_radius: float,
        env_size: float,
    ) -> None:
        """
        Default predator-prey observation function

        Produces an individual local view of other agents in the
        environment for each agent. Each agents view is a
        ``[2, num_vision]`` array, where the values represent the
        distance along a ray to the nearest agent, with -1
        representing the case no agent is in range. The two
        rows represent the view of different types of agents.

        Parameters
        ----------
        num_vision
            Number of cells/values in the segmented view
        predator_vision_range
            Predator agent vision range
        prey_vision_range
            Prey vision range
        predator_view_angle
            Predator view angle, as a fraction of π
        prey_view_angle
            Prey view angle, as a fraction of π
        agent_radius
            Agent visual radius
        env_size
            Environment size
        """
        self.num_vision = num_vision
        self.predator_vision_range = predator_vision_range
        self.prey_vision_range = prey_vision_range
        self.predator_view_angle = predator_view_angle
        self.prey_view_angle = prey_view_angle
        self.agent_radius = agent_radius
        self.env_size = env_size

    def __call__(self, state: State) -> tuple[chex.Array, chex.Array]:
        """
        Generate predator and prey views from the current state

        Parameters
        ----------
        state
            Environment state

        Returns
        -------
        Array, Array
            Tuple containing predator and prey views respectively
        """
        prey_view_predators = esquilax.transforms.spatial(
            view,
            reduction=_view_reduction((self.num_vision,)),
            include_self=False,
            i_range=self.prey_vision_range,
            dims=self.env_size,
        )(
            (self.prey_view_angle, self.agent_radius),
            state.prey,
            state.predators,
            pos=state.prey.pos,
            pos_b=state.predators.pos,
            n_view=self.num_vision,
            i_range=self.prey_vision_range,
            env_size=self.env_size,
        )
        prey_view_prey = esquilax.transforms.spatial(
            view,
            reduction=_view_reduction((self.num_vision,)),
            include_self=False,
            i_range=self.prey_vision_range,
            dims=self.env_size,
        )(
            (self.prey_view_angle, self.agent_radius),
            state.prey,
            state.prey,
            pos=state.prey.pos,
            n_view=self.num_vision,
            i_range=self.prey_vision_range,
            env_size=self.env_size,
        )
        predators_view_prey = esquilax.transforms.spatial(
            view,
            reduction=_view_reduction((self.num_vision,)),
            include_self=False,
            i_range=self.predator_vision_range,
            dims=self.env_size,
        )(
            (self.predator_view_angle, self.agent_radius),
            state.predators,
            state.prey,
            pos=state.predators.pos,
            pos_b=state.prey.pos,
            n_view=self.num_vision,
            i_range=self.predator_vision_range,
            env_size=self.env_size,
        )
        predators_view_predators = esquilax.transforms.spatial(
            view,
            reduction=_view_reduction((self.num_vision,)),
            include_self=False,
            i_range=self.predator_vision_range,
            dims=self.env_size,
        )(
            (self.predator_view_angle, self.agent_radius),
            state.predators,
            state.predators,
            pos=state.predators.pos,
            n_view=self.num_vision,
            i_range=self.predator_vision_range,
            env_size=self.env_size,
        )

        prey_views = jnp.stack((prey_view_prey, prey_view_predators), axis=1)
        predator_views = jnp.stack(
            (predators_view_predators, predators_view_prey), axis=1
        )

        return predator_views, prey_views



"""Predator-prey environment reward functions"""
import abc

import esquilax
from jumanji.environments.swarms.common.types import AgentState

from .types import Rewards, State


class RewardFn(abc.ABC):
    """Predator-prey reward function interface"""

    @abc.abstractmethod
    def __call__(self, state: State) -> Rewards:
        """Generate predator and prey rewards from state

        Parameters
        ----------
        state
            Env state

        Returns
        -------
        Rewards
            Individual reward for each predator and prey agent
        """


class SparseRewards(RewardFn):
    def __init__(
        self,
        capture_radius: float,
        predator_reward: float,
        prey_penalty: float,
    ) -> None:
        """
        Sparse reward function

        Sparse reward function that provides a fixed reward/penalty
        to predators/prey when in capture range, irrespective of
        distance.

        Predator agents are provided a fixed reward if within capture range
        of a prey agent. Prey agents are provide a fixed reward summed
        over all prey agents within capture range.

        Parameters
        ----------
        capture_radius
            Range from predators at which prey are considered captured
        predator_reward
            Fixed reward provided to predators
        prey_penalty
            Fixed penalty provided to prey
        """
        self.capture_radius = capture_radius
        self.predator_reward = predator_reward
        self.prey_penalty = prey_penalty

    def __call__(self, state: State) -> Rewards:
        """
        Generate rewards

        Parameters
        ----------
        state
            Current environment state

        Returns
        -------
        Rewards
            Struct containing predator and prey reward arrays
        """
        prey_rewards = esquilax.transforms.spatial(
            lambda reward, _a, _b: -reward,
            reduction=esquilax.reductions.add(),
            include_self=False,
            i_range=self.capture_radius,
        )(
            self.prey_penalty,
            None,
            None,
            pos=state.prey.pos,
            pos_b=state.predators.pos,
        )
        predator_rewards = esquilax.transforms.nearest_neighbour(
            lambda reward, _a, _b: reward,
            default=0.0,
            i_range=self.capture_radius,
        )(
            self.predator_reward,
            None,
            None,
            pos=state.predators.pos,
            pos_b=state.prey.pos,
        )
        return Rewards(
            predators=predator_rewards,
            prey=prey_rewards,
        )


def _distance_reward(
    reward: float, a: AgentState, b: AgentState, *, i_range, env_size
) -> float:
    d = esquilax.utils.shortest_distance(a.pos, b.pos, env_size) / i_range
    return reward * (d - 1.0)


class LinearRewards(RewardFn):
    def __init__(
        self,
        capture_radius: float,
        predator_reward: float,
        prey_penalty: float,
        env_size: float,
    ) -> None:
        """
        Rewards function that scales reward by linear distance

        Reward function that assigns rewards/penalties to
        predator/prey if within capture range. The rewards
        are scaled linearly based on the distance between the
        predator and prey.

        Predator agents are provided a reward if within capture range
        of a prey agent. Prey agents are provide a reward summed
        over all prey agents within capture range.

        Parameters
        ----------
        capture_radius
            Agent capture radius
        predator_reward
            Max predator reward
        prey_penalty
            Max prey penalty
        env_size
            Environment size
        """
        self.capture_radius = capture_radius
        self.predator_reward = predator_reward
        self.prey_penalty = prey_penalty
        self.env_size = env_size

    def __call__(self, state: State) -> Rewards:
        """
        Generate rewards

        Parameters
        ----------
        state
            Current environment state

        Returns
        -------
        Rewards
            Struct containing predator and prey reward arrays
        """
        prey_rewards = esquilax.transforms.spatial(
            _distance_reward,
            reduction=esquilax.reductions.add(),
            include_self=False,
            i_range=2 * self.capture_radius,
        )(
            -self.prey_penalty,
            None,
            None,
            pos=state.prey.pos,
            pos_b=state.predators.pos,
            i_range=2 * self.capture_radius,
            env_size=self.env_size,
        )
        predator_rewards = esquilax.transforms.nearest_neighbour(
            _distance_reward,
            default=0.0,
            i_range=2 * self.capture_radius,
        )(
            self.predator_reward,
            None,
            None,
            pos=state.predators.pos,
            pos_b=state.prey.pos,
            i_range=2 * self.capture_radius,
            env_size=self.env_size,
        )
        return Rewards(
            predators=predator_rewards,
            prey=prey_rewards,
        )



"""Flock environment initial state generator"""
import abc

import chex
from jumanji.environments.swarms.common.types import AgentParams

from floxs.common.utils import random_agent_state

from .types import State


class Generator(abc.ABC):
    def __init__(self, num_boids: int, env_size: float = 1.0) -> None:
        """
        Abstract flock environment state generator

        Parameters
        ----------
        num_boids
            Number of agents in the flock
        env_size
            Size of the environment, default ``1.0``
        """
        self.num_boids = num_boids
        self.env_size = env_size

    @abc.abstractmethod
    def __call__(self, key: chex.PRNGKey, boid_params: AgentParams) -> State:
        """
        Generate initial state

        Parameters
        ----------
        key
            JAX random key
        boid_params
            Boid agent parameters

        Returns
        -------
        State
            Initial environment state
        """
        """Generate initial agent positions and velocities.

        Args:
            key: JAX random key
            boid_params: Agent `AgentParams` parameters

        Returns:
            Initial agent `State`.
        """


class RandomGenerator(Generator):
    def __call__(self, key: chex.PRNGKey, boid_params: AgentParams) -> State:
        """
        Random generator, sampling from a uniform distribution

        Parameters
        ----------
        key
            JAX random key
        boid_params
            Boid agent parameters

        Returns
        -------
        State
            Randomly sampled initial state
        """
        boid_states = random_agent_state(
            key, boid_params, self.num_boids, self.env_size
        )
        state = State(
            boids=boid_states,
        )
        return state



"""
Predator-prey multi-agent environment

Multi-agent environment where a Flock of predator
type agents attempt to capture a flock of prey
agents.
"""
from functools import cached_property
from typing import Optional, Sequence

import chex
import jax
import jax.numpy as jnp
from jumanji import specs
from jumanji.env import Environment
from jumanji.environments.swarms.common.types import AgentParams
from jumanji.environments.swarms.common.updates import update_state
from jumanji.types import StepType, TimeStep
from jumanji.viewer import Viewer
from matplotlib.animation import FuncAnimation

from .generator import Generator, RandomGenerator
from .observations import ObservationFn
from .rewards import RewardFn, SparseRewards
from .types import Actions, Discounts, Observation, Rewards, State
from .viewer import PredatorPreyViewer


class PredatorPrey(Environment):
    def __init__(
        self,
        prey_max_rotate: float = 0.025,
        prey_max_accelerate: float = 0.001,
        prey_min_speed: float = 0.015,
        prey_max_speed: float = 0.025,
        predator_max_rotate: float = 0.025,
        predator_max_accelerate: float = 0.001,
        predator_min_speed: float = 0.015,
        predator_max_speed: float = 0.025,
        time_limit: int = 400,
        viewer: Optional[Viewer[State]] = None,
        generator: Optional[Generator] = None,
        reward_fn: Optional[RewardFn] = None,
        observation: Optional[ObservationFn] = None,
    ) -> None:
        """
        Predator-prey multi-agent RL environment

        Environment containing two distinct agent types, predators and prey.
        The predator agents are rewarded for coming within capture range of
        the prey agents. Conversely, the prey agents are penalised if within
        capture range of a prey agent.

        Actions
        -------
        Each set of can take the same actions, changing heading, and changing
        speed. Actions are in the range ``[-1, 1]`` which are then
        scales by the maximum rotation and acceleration parameters for the
        respective agent types. The total actions for the agent types are are
        arrays  with shape ``[n_predators, 2]`` and ``[n_prey, 2]. The action
        values represent  ``[rotation, acceleration]`` respectively.

        Observations
        ------------
        All agents individually observe their local environment within a
        given range. The view for each agent is an array of shape
        ``[2, n_vision]``. Each entry represents the distance along a ray
        to the nearest neighbouring agent, where the default ``-1`` represents
        a lock of agent along that ray. The 2 rows of the array are used
        to represent the two distinct agent types.

        Rewards
        -------
        By default the rewards are:

        - Predator agents are provided a fixed reward when within capture
          range of a prey agent
        - Prey agents are penalised for every predator agent within capture range

        Parameters
        ----------
        prey_max_rotate
            Max prey agent change in heading in a single
            step, as a fraction of π
        prey_max_accelerate
            Max prey agent change in speed in a single
            step, as a fraction of π
        prey_min_speed
            Prey agent minimum speed
        prey_max_speed
            Prey agent maximum speed
        predator_max_rotate
            Max predator agent change in heading in a single
            step, as a fraction of π
        predator_max_accelerate
            Max predator agent change in speed in a single
            step, as a fraction of π
        predator_min_speed
            Predator agent minimum speed
        predator_max_speed
            Predator agent maximum speed
        time_limit
            Environment time limit
        viewer
            Plot/image generator, default uses the Matplotlib backend
        generator
            Initial state generator, default generates a uniform random
            distribution of predator and prey agents
        reward_fn
            Reward function, default provides a reward of 0.1 to
            predators if within range of a prey agent, and a -0.1
            penalty to prey for each predator within capture range
        observation
            Agent observation function, default generates a segmented
            view with 2 channels, containing the distance to agents
            of the same-type, and opposite type in different channels
        """
        self.predator_params = AgentParams(
            max_rotate=predator_max_rotate,
            max_accelerate=predator_max_accelerate,
            min_speed=predator_min_speed,
            max_speed=predator_max_speed,
        )
        self.prey_params = AgentParams(
            max_rotate=prey_max_rotate,
            max_accelerate=prey_max_accelerate,
            min_speed=prey_min_speed,
            max_speed=prey_max_speed,
        )
        self.time_limit = time_limit
        self.generator = generator or RandomGenerator(num_predators=2, num_prey=20)
        self._viewer = viewer or PredatorPreyViewer()
        self._reward_fn = reward_fn or SparseRewards(
            capture_radius=0.02,
            predator_reward=0.1,
            prey_penalty=0.1,
        )
        self._observation_fn = observation or ObservationFn(
            num_vision=128,
            predator_vision_range=0.2,
            prey_vision_range=0.1,
            predator_view_angle=0.25,
            prey_view_angle=0.6,
            env_size=self.generator.env_size,
            agent_radius=0.01,
        )
        super().__init__()

    def __repr__(self) -> str:
        predator_vision_range = self._observation_fn.predator_vision_range
        return "\n".join(
            [
                "Boid flock multi-agent environment:",
                f" - num predators: {self.generator.num_predators}",
                f" - num prey: {self.generator.num_prey}",
                f" - predator max rotation: {self.predator_params.max_rotate}",
                f" - predator max acceleration: {self.predator_params.max_accelerate}",
                f" - predator min speed: {self.predator_params.min_speed}",
                f" - predator max speed: {self.predator_params.max_speed}",
                f" - prey max rotation: {self.prey_params.max_rotate}",
                f" - prey max acceleration: {self.prey_params.max_accelerate}",
                f" - prey min speed: {self.prey_params.min_speed}",
                f" - prey max speed: {self.prey_params.max_speed}",
                f" - predator vision range: {predator_vision_range}",
                f" - prey vision range: {self._observation_fn.prey_vision_range}",
                f" - predator view angle: {self._observation_fn.predator_view_angle}",
                f" - prey view angle: {self._observation_fn.prey_view_angle}",
                f" - num vision: {self._observation_fn.num_vision}",
                f" - agent radius: {self._observation_fn.agent_radius}",
                f" - time limit: {self.time_limit},"
                f" - env size: {self.generator.env_size}"
                f" - generator: {self.generator.__class__.__name__}",
                f" - reward fn: {self._reward_fn.__class__.__name__}",
                f" - observation fn: {self._observation_fn.__class__.__name__}",
            ]
        )

    def reset(self, key: chex.PRNGKey) -> tuple[State, TimeStep[Observation]]:
        """
        Reset the environment

        Generates a new initial environment state and initial timestep

        Parameters
        ----------
        key
            JAX random key

        Returns
        -------
        State, TimeStep
            New initial state and initial timestep
        """
        state = self.generator(key, self.predator_params, self.prey_params)
        timestep = TimeStep(
            step_type=StepType.FIRST,
            reward=Rewards(
                predators=jnp.zeros(self.generator.num_predators, dtype=float),
                prey=jnp.zeros(self.generator.num_prey, dtype=float),
            ),
            discount=Discounts(
                predators=jnp.ones(self.generator.num_predators, dtype=float),
                prey=jnp.ones(self.generator.num_prey, dtype=float),
            ),
            observation=self._state_to_observation(state),
            extras={},
        )
        return state, timestep

    def step(
        self, state: State, actions: Actions
    ) -> tuple[State, TimeStep[Observation]]:
        """
        Apply actions and update the state of the environment

        The environment is updated in several steps:

        - Apply predator and prey actions
        - Update predator and prey positions
        - Calculate new individual rewards for each agent type
        - Generate individual views for each agent and type

        Parameters
        ----------
        state
            Current environment state
        actions
            Actions struct containing predator and prey action arrays,
            in shapes ``[n-predators, 2]`` and ``[n-prey, 2]``
            respectively. Actions lie in the range ``[-1, 1]`` which
            are then scaled by the corresponding parameters.
            The 2 values represent ``[rotation, acceleration]``.

        Returns
        -------
        State, TimeStep
            Updated environment state and new timestep
        """
        predators = update_state(
            self.generator.env_size,
            self.predator_params,
            state.predators,
            actions.predators,
        )
        prey = update_state(
            self.generator.env_size, self.prey_params, state.prey, actions.prey
        )
        state = State(predators=predators, prey=prey, step=state.step + 1)
        rewards = self._reward_fn(state)
        observation = self._state_to_observation(state)
        observation = jax.lax.stop_gradient(observation)
        step_type, discounts = jax.lax.cond(
            state.step >= self.time_limit,
            lambda: (
                StepType.LAST,
                Discounts(
                    predators=jnp.zeros(self.generator.num_predators, dtype=float),
                    prey=jnp.zeros(self.generator.num_prey, dtype=float),
                ),
            ),
            lambda: (
                StepType.MID,
                Discounts(
                    predators=jnp.ones(self.generator.num_predators, dtype=float),
                    prey=jnp.ones(self.generator.num_prey, dtype=float),
                ),
            ),
        )
        timestep = TimeStep(
            step_type=step_type,
            reward=rewards,
            discount=discounts,
            observation=observation,
            extras={},
        )
        return state, timestep

    def _state_to_observation(self, state: State) -> Observation:
        predator_views, prey_views = self._observation_fn(state)
        return Observation(
            predator_views=predator_views,
            prey_views=prey_views,
            step=state.step,
            predator_positions=state.predators.pos / self.generator.env_size,
            prey_positions=state.prey.pos / self.generator.env_size,
        )

    @cached_property
    def observation_spec(self) -> specs.Spec[Observation]:
        """Returns the observation spec

        Local predator and prey agent views representing the distance to the
        closest neighbouring agents of each type

        Returns
        -------
        Spec
            Predator-prey environment observation spec
        """
        predator_views = specs.BoundedArray(
            shape=(
                self.generator.num_predators,
                2,
                self._observation_fn.num_vision,
            ),
            minimum=-1.0,
            maximum=1.0,
            dtype=float,
            name="predator_view",
        )
        prey_views = specs.BoundedArray(
            shape=(
                self.generator.num_prey,
                2,
                self._observation_fn.num_vision,
            ),
            minimum=-1.0,
            maximum=1.0,
            dtype=float,
            name="prey_view",
        )
        return specs.Spec(
            Observation,
            "ObservationSpec",
            predator_views=predator_views,
            prey_views=prey_views,
            step=specs.BoundedArray(
                shape=(),
                minimum=0,
                maximum=self.time_limit,
                name="step",
                dtype=jnp.int32,
            ),
            predator_positions=specs.BoundedArray(
                shape=(self.generator.num_predators, 2),
                minimum=0.0,
                maximum=1.0,
                name="positions",
                dtype=jnp.float32,
            ),
            prey_positions=specs.BoundedArray(
                shape=(self.generator.num_prey, 2),
                minimum=0.0,
                maximum=1.0,
                name="positions",
                dtype=jnp.float32,
            ),
        )

    @cached_property
    def action_spec(self) -> specs.Spec[Actions]:
        """Returns the action spec

        Actions struct containing predator and prey action arrays

        Returns
        -------
        Spec
            Action array spec
        """
        return specs.Spec(
            Actions,
            predators=specs.BoundedArray(
                shape=(self.generator.num_predators, 2),
                minimum=-1.0,
                maximum=1.0,
                dtype=float,
            ),
            prey=specs.BoundedArray(
                shape=(self.generator.num_prey, 2),
                minimum=-1.0,
                maximum=1.0,
                dtype=float,
            ),
        )

    @cached_property
    def reward_spec(self) -> specs.Spec[Rewards]:
        """Returns the reward spec

        Rewards struct, containing individual rewards for
        predator and prey agents

        Returns
        -------
        Spec
            Reward spec
        """
        return specs.Spec(
            Rewards,
            predators=specs.Array(
                shape=(self.generator.num_predators,),
                dtype=float,
            ),
            prey=specs.Array(
                shape=(self.generator.num_prey,),
                dtype=float,
            ),
        )

    def render(self, state: State) -> None:
        """Render a frame of the environment for a given state using matplotlib

        Parameters
        ----------
        state
            State object
        """
        self._viewer.render(state)

    def animate(
        self,
        states: Sequence[State],
        interval: int = 100,
        save_path: Optional[str] = None,
    ) -> FuncAnimation:
        """Create an animation from a sequence of environment states

        Parameters
        ----------
        states
            Sequence of environment states corresponding to consecutive
            timesteps
        interval
            Delay between frames in milliseconds.
        save_path
            The path where the animation file should be saved. If it
            is None, the plot will not be saved

        Returns
        -------
        FuncAnimation
            Animation that can be saved as a GIF, MP4, or rendered with HTML
        """
        return self._viewer.animate(states, interval=interval, save_path=save_path)

    def close(self) -> None:
        """Perform any necessary cleanup."""
        self._viewer.close()



"""Flock agent reward functions"""
import abc
from typing import Callable, Union

import chex
import esquilax
import jax
import jax.numpy as jnp
from jumanji.environments.swarms.common.types import AgentState

from .types import State


def _apply_reward_fn(
    boid_radius: float,
    boid_a: AgentState,
    boid_b: AgentState,
    *,
    f: Callable,
    i_range: float,
) -> chex.Array:
    d = esquilax.utils.shortest_distance(boid_a.pos, boid_b.pos, norm=True)
    reward = f(d / i_range)
    return jax.lax.cond(d < 2 * boid_radius, lambda: (1, reward), lambda: (0, reward))


class RewardFn(abc.ABC):
    """
    Reward function interface
    """

    @abc.abstractmethod
    def __call__(self, state: State) -> chex.Array:
        """Generate rewards from state

        Parameters
        ----------
        state
            Env state

        Returns
        -------
        Array
            Array of individual reward for each agent
        """


class BaseDistanceRewardFn(RewardFn):
    def __init__(
        self,
        boid_radius: float,
        collision_penalty: float,
        i_range: float,
        reward_fn: Callable[[float], Union[float, chex.Array]],
    ) -> None:
        """
        Base distance based reward function

        Base reward function that generated rewards based on the
        distance between individual agents (up to a given range).

        A negative penalty is applied if the agents collide.

        Rewards are accumulated between all pairs of agents within range.

        Parameters
        ----------
        boid_radius
            Radius of agents
        collision_penalty
            Penalty returned in case of colliding agents
        i_range
            Interaction range
        reward_fn
            Distance reward function
        """
        self.boid_radius = boid_radius
        self.collision_penalty = collision_penalty
        self.i_range = i_range
        self.reward_fn = reward_fn

    def __call__(self, state: State) -> chex.Array:
        """Generate rewards from state

        Parameters
        ----------
        state
            Env state

        Returns
        -------
        Array
            Array of individual reward for each agent
        """
        collisions, rewards = esquilax.transforms.spatial(
            _apply_reward_fn,
            reduction=esquilax.reductions.Reduction((jnp.add, jnp.add), (0, 0.0)),
            include_self=False,
            topology="moore",
            i_range=self.i_range,
        )(
            self.boid_radius,
            state.boids,
            state.boids,
            pos=state.boids.pos,
            f=self.reward_fn,
            i_range=self.i_range,
        )
        rewards = jnp.where(collisions > 0, -self.collision_penalty, rewards)
        return rewards


def _exponential_rewards(d: float) -> chex.Array:
    return jnp.exp(-5 * d)


class ExponentialRewardFn(BaseDistanceRewardFn):
    def __init__(
        self,
        boid_radius: float,
        collision_penalty: float,
        i_range: float,
    ) -> None:
        """Rewards that decrease exponentially with distance

        Parameters
        ----------
        boid_radius
            Radius of agents
        collision_penalty
            Penalty returned in case of colliding agents
        i_range
            Interaction range
        """
        super().__init__(boid_radius, collision_penalty, i_range, _exponential_rewards)



"""Flock agent observation function"""
from typing import Tuple

import chex
import esquilax
import jax.numpy as jnp
from jumanji.environments.swarms.common.updates import view, view_reduction_fn

from .types import State


def _view_reduction(view_shape: Tuple[int, ...]) -> esquilax.reductions.Reduction:
    return esquilax.reductions.Reduction(
        fn=view_reduction_fn,
        id=-jnp.ones(view_shape),
    )


class ObservationFn:
    def __init__(
        self,
        num_vision: int,
        vision_range: float,
        view_angle: float,
        boid_radius: float,
        env_size: float,
    ) -> None:
        """
        Base agent environment observation function

        Parameters
        ----------
        num_vision
            Number of segments/cells in the observation
        vision_range
            Vision of range of individual agents
        view_angle
            View angle of individual agents (as a fraction
            of pi from the agents heading)
        boid_radius
            Visual radius of the agents
        env_size
            Size of the environment
        """
        self.num_vision = num_vision
        self.vision_range = vision_range
        self.view_angle = view_angle
        self.boid_radius = boid_radius
        self.env_size = env_size

    def __call__(self, state: State) -> chex.Array:
        """
        Generate agent observations

        Parameters
        ----------
        state
            Current environment state

        Returns
        -------
        Array
            Array of individual agent observation arrays, in
            shape ``[n-agents, n-observation]``
        """
        views = esquilax.transforms.spatial(
            view,
            reduction=_view_reduction((self.num_vision,)),
            include_self=False,
            i_range=self.vision_range,
            dims=self.env_size,
        )(
            (self.view_angle, self.boid_radius),
            state.boids,
            state.boids,
            pos=state.boids.pos,
            n_view=self.num_vision,
            i_range=self.vision_range,
            env_size=self.env_size,
        )
        return views



"""Flock environment data types"""
from typing import TYPE_CHECKING, NamedTuple

import chex
from jumanji.environments.swarms.common.types import AgentState

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from chex import dataclass


@dataclass
class State:
    """
    Environment state

    boids
        State of flock agents
    step
        Current simulation step
    """

    boids: AgentState
    step: int = 0


class Observation(NamedTuple):
    """
    Agent observations

    views
        Array of individual agent views on the environment
    step
        Current environment step
    positions
        Positions of individual agents in the environment
    """

    views: chex.Array  # (num_boids, num_vision)
    step: chex.Numeric  # ()
    positions: chex.Array  # (num_boids, 2)
